import os
import warnings
import pytest
import numpy as np
import yaml
from contextlib import contextmanager
# import pkg_resources
from pyrealm import pmodel


# RPMODEL bugs
# rpmodel was using an incorrect parameterisation of the C4 ftemp kphio curve
# that is fixed but currently (1.2.0) an implementation error in the output checking 
# means this still have to be skipped.

RPMODEL_C4_BUG = True

# ------------------------------------------
# Null context manager to include exception testing in test paramaterisation
# ------------------------------------------

@contextmanager
def does_not_raise():
    yield

# ------------------------------------------
# Fixtures: inputs and expected values
# ------------------------------------------


@pytest.fixture(scope='module')
def values():
    """Fixture to load test inputs from file.
    """

    test_dir = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(test_dir, 'test_outputs_rpmodel.yaml')) as infile:
        values = yaml.load(infile, Loader=yaml.SafeLoader)

    values = {k: np.array(v) if isinstance(v, list) else v
              for k, v in values.items()}

    return values

# ------------------------------------------
# Test structure
# The basic structure of these tests  is to use a pytest.mark.parameterise
# to pass in a dictionary of the variables to be passed to the function and then
# a dictionary identifying the expected values that have been read in from file.
#
# A separate R file is used to read in the same inputs to R and run them through
# rpmodel to generate the output file.
#
# TODO - should be able to group shared inputs (e.g. tc + patm) to recycle
#        parameterisation.
# ------------------------------------------

# ------------------------------------------
# Testing calc_density_h20 - temp + patm
# ------------------------------------------


@pytest.mark.parametrize(
    'ctrl',
    [(dict(args=dict(tc='tc_sc', patm='patm_sc'),  # scalars
           cmng=does_not_raise(),
           out='dens_h20_sc')),
     (dict(args=dict(tc='tc_ar', patm='patm_sc'),  # scalar + array
           cmng=does_not_raise(),
           out='dens_h20_mx')),
     (dict(args=dict(tc='tc_ar', patm='patm_ar'),  # arrays
           cmng=does_not_raise(),
           out='dens_h20_ar')),
     (dict(args=dict(tc='tc_ar', patm='shape_error'),  # shape mismatch
           cmng=pytest.raises(ValueError),
           out=None))]
)
def test_calc_density_h2o(values, ctrl):

    with ctrl['cmng']:
        kwargs = {k: values[v] for k, v in ctrl['args'].items()}
        ret = pmodel.calc_density_h2o(**kwargs)
        assert np.allclose(ret, values[ctrl['out']])

# ------------------------------------------
# Testing calc_ftemp_inst_rd - temp only but in Kelvin!
# ------------------------------------------


@pytest.mark.parametrize(
    'ctrl',
    [(dict(args=dict(tk='tk_sc', ha='KattgeKnorr_ha'),  # scalar
           cmng=does_not_raise(),
           out='ftemp_arrh_sc')),
     (dict(args=dict(tk='tk_ar', ha='KattgeKnorr_ha'),  # array
           cmng=does_not_raise(),
           out='ftemp_arrh_ar'))]
)
def test_calc_ftemp_arrh(values, ctrl):

    with ctrl['cmng']:
        kwargs = {k: values[v] for k, v in ctrl['args'].items()}
        ret = pmodel.calc_ftemp_arrh(**kwargs)
        assert np.allclose(ret, values[ctrl['out']])

# ------------------------------------------
# Testing calc_ftemp_inst_vcmax - temp only
# ------------------------------------------


@pytest.mark.parametrize(
    'ctrl',
    [(dict(args=dict(tc='tc_sc'),  # scalar
           cmng=does_not_raise(),
           out='ftemp_inst_vcmax_sc')),
     (dict(args=dict(tc='tc_ar'),  # array
           cmng=does_not_raise(),
           out='ftemp_inst_vcmax_ar'))]
)
def test_calc_ftemp_inst_vcmax(values, ctrl):

    with ctrl['cmng']:
        kwargs = {k: values[v] for k, v in ctrl['args'].items()}
        ret = pmodel.calc_ftemp_inst_vcmax(**kwargs)
        assert np.allclose(ret, values[ctrl['out']])

# ------------------------------------------
# Testing calc_ftemp_inst_vcmax - temp only
# ------------------------------------------


@pytest.mark.skipif(RPMODEL_C4_BUG, reason='Benchmark incorrect')
@pytest.mark.parametrize(
    'ctrl',
    [(dict(args=dict(tc='tc_sc'), c4=False,  # scalar, C3
           cmng=does_not_raise(),
           out='ftemp_kphio_c3_sc')),
     (dict(args=dict(tc='tc_ar'), c4=False,  # array, C3
           cmng=does_not_raise(),
           out='ftemp_kphio_c3_ar')),
     (dict(args=dict(tc='tc_sc'), c4=True,  # scalar, C4
           cmng=does_not_raise(),
           out='ftemp_kphio_c4_sc')),
     (dict(args=dict(tc='tc_ar'), c4=True,  # array, C4
           cmng=does_not_raise(),
           out='ftemp_kphio_c4_ar'))]
)
def test_calc_ftemp_kphio(values, ctrl):

    with ctrl['cmng']:
        kwargs = {k: values[v] for k, v in ctrl['args'].items()}
        ret = pmodel.calc_ftemp_kphio(**kwargs, c4=ctrl['c4'])
        assert np.allclose(ret, values[ctrl['out']])

# ------------------------------------------
# Testing calc_gammastar - temp + patm
# ------------------------------------------


@pytest.mark.parametrize(
    'ctrl',
    [(dict(args=dict(tc='tc_sc', patm='patm_sc'),  # scalars
           cmng=does_not_raise(),
           out='gammastar_sc')),
     (dict(args=dict(tc='tc_ar', patm='patm_sc'),  # scalar + array
           cmng=does_not_raise(),
           out='gammastar_mx')),
     (dict(args=dict(tc='tc_ar', patm='patm_ar'),  # arrays
           cmng=does_not_raise(),
           out='gammastar_ar')),
     (dict(args=dict(tc='tc_ar', patm='shape_error'),  # shape mismatch
           cmng=pytest.raises(ValueError),
           out=None))]
)
def test_calc_gammastar(values, ctrl):

    with ctrl['cmng']:
        kwargs = {k: values[v] for k, v in ctrl['args'].items()}
        ret = pmodel.calc_gammastar(**kwargs)
        assert np.allclose(ret, values[ctrl['out']])

# ------------------------------------------
# Testing calc_kmm - temp + patm
# ------------------------------------------


@pytest.mark.parametrize(
    'ctrl',
    [(dict(args=dict(tc='tc_sc', patm='patm_sc'),  # scalars
           cmng=does_not_raise(),
           out='kmm_sc')),
     (dict(args=dict(tc='tc_ar', patm='patm_sc'),  # scalar + array
           cmng=does_not_raise(),
           out='kmm_mx')),
     (dict(args=dict(tc='tc_ar', patm='patm_ar'),  # arrays
           cmng=does_not_raise(),
           out='kmm_ar')),
     (dict(args=dict(tc='tc_ar', patm='shape_error'),  # shape mismatch
           cmng=pytest.raises(ValueError),
           out=None))]
)
def test_calc_kmm(values, ctrl):

    with ctrl['cmng']:
        kwargs = {k: values[v] for k, v in ctrl['args'].items()}
        ret = pmodel.calc_kmm(**kwargs)
        assert np.allclose(ret, values[ctrl['out']])

# ------------------------------------------
# Testing calc_soilmstress - soilm + meanalpha
# ------------------------------------------


@pytest.mark.parametrize(
    'ctrl',
    [(dict(args=dict(soilm='soilm_sc', meanalpha='meanalpha_sc'),  # scalars
           cmng=does_not_raise(),
           out='soilmstress_sc')),
     (dict(args=dict(soilm='soilm_ar', meanalpha='meanalpha_sc'),  # scalar + array
           cmng=does_not_raise(),
           out='soilmstress_mx')),
     (dict(args=dict(soilm='soilm_ar', meanalpha='meanalpha_ar'),  # arrays
           cmng=does_not_raise(),
           out='soilmstress_ar')),
     (dict(args=dict(soilm='soilm_ar', meanalpha='shape_error'),  # shape mismatch
           cmng=pytest.raises(ValueError),
           out=None))]
)
def test_calc_soilmstress(values, ctrl):

    with ctrl['cmng']:
        kwargs = {k: values[v] for k, v in ctrl['args'].items()}
        ret = pmodel.calc_soilmstress(**kwargs)
        assert np.allclose(ret, values[ctrl['out']])

# ------------------------------------------
# Testing calc_viscosity_h2o - temp + patm
# ------------------------------------------


@pytest.mark.parametrize(
    'ctrl',
    [(dict(args=dict(tc='tc_sc', patm='patm_sc'),  # scalars
           cmng=does_not_raise(),
           out='viscosity_h2o_sc')),
     (dict(args=dict(tc='tc_ar', patm='patm_sc'),  # scalar + array
           cmng=does_not_raise(),
           out='viscosity_h2o_mx')),
     (dict(args=dict(tc='tc_ar', patm='patm_ar'),  # arrays
           cmng=does_not_raise(),
           out='viscosity_h2o_ar')),
     (dict(args=dict(tc='tc_ar', patm='shape_error'),  # shape mismatch
           cmng=pytest.raises(ValueError),
           out=None))]
)
def test_calc_viscosity_h2o(values, ctrl):

    with ctrl['cmng']:
        kwargs = {k: values[v] for k, v in ctrl['args'].items()}
        ret = pmodel.calc_viscosity_h2o(**kwargs)
        assert np.allclose(ret, values[ctrl['out']])

# ------------------------------------------
# Testing calc_patm - elev only
# ------------------------------------------


@pytest.mark.parametrize(
    'ctrl',
    [(dict(args=dict(elv='elev_sc'),  # scalar
           cmng=does_not_raise(),
           out='patm_from_elev_sc')),
     (dict(args=dict(elv='elev_ar'),  # array
           cmng=does_not_raise(),
           out='patm_from_elev_ar'))]
)
def test_calc_patm(values, ctrl):

    with ctrl['cmng']:
        kwargs = {k: values[v] for k, v in ctrl['args'].items()}
        ret = pmodel.calc_patm(**kwargs)
        assert np.allclose(ret, values[ctrl['out']])

# ------------------------------------------
# Testing calc_co2_to_ca - co2 + patm
# ------------------------------------------


@pytest.mark.parametrize(
    'ctrl',
    [(dict(args=dict(co2='co2_sc', patm='patm_sc'),  # scalars
           cmng=does_not_raise(),
           out='ca_sc')),
     (dict(args=dict(co2='co2_ar', patm='patm_sc'),  # scalar + array
           cmng=does_not_raise(),
           out='ca_mx')),
     (dict(args=dict(co2='co2_ar', patm='patm_ar'),  # arrays
           cmng=does_not_raise(),
           out='ca_ar')),
     (dict(args=dict(co2='co2_ar', patm='shape_error'),  # shape mismatch
           cmng=pytest.raises(ValueError),
           out=None))]
)
def test_calc_co2_to_ca(values, ctrl):

    with ctrl['cmng']:
        kwargs = {k: values[v] for k, v in ctrl['args'].items()}
        ret = pmodel.calc_co2_to_ca(**kwargs)
        assert np.allclose(ret, values[ctrl['out']])


# ------------------------------------------
# Testing CalcOptimalChi - vpd + internals kmm, gammastar, ns_star, ca
#
# NOTE - the c4 method __INTENTIONALLY__ always returns scalars
#        regardless of input shape, so always uses the same expected values
# ------------------------------------------


@pytest.mark.parametrize(
    'ctrl',
    [(dict(args=dict(kmm='kmm_sc', gammastar='gammastar_sc',
                     ns_star='ns_star_sc', ca='ca_sc', vpd='vpd_sc'),
           method='c4',  # scalar, C4
           cmng=does_not_raise(),
           out='optchi_c4')),
     (dict(args=dict(kmm='kmm_ar', gammastar='gammastar_sc',
                     ns_star='ns_star_ar', ca='ca_sc', vpd='vpd_sc'),
           method='c4',  # scalar + arrays, C4
           cmng=does_not_raise(),
           out='optchi_c4')),
     (dict(args=dict(kmm='kmm_ar', gammastar='gammastar_ar',
                     ns_star='ns_star_ar', ca='ca_ar', vpd='vpd_ar'),
           method='c4',  # scalar + arrays, C4
           cmng=does_not_raise(),
           out='optchi_c4')),
     (dict(args=dict(kmm='kmm_ar', gammastar='shape_error',
                     ns_star='ns_star_ar', ca='ca_ar', vpd='vpd_ar'),
           method='c4',  # scalar + arrays, C4
           cmng=pytest.raises(ValueError),
           out=None)),
     (dict(args=dict(kmm='kmm_sc', gammastar='gammastar_sc',
                     ns_star='ns_star_sc', ca='ca_sc', vpd='vpd_sc'),
           method='prentice14',  # scalar, c3
           cmng=does_not_raise(),
           out='optchi_p14_sc')),
     (dict(args=dict(kmm='kmm_ar', gammastar='gammastar_sc',
                     ns_star='ns_star_ar', ca='ca_sc', vpd='vpd_sc'),
           method='prentice14',  # scalar + arrays, c3
           cmng=does_not_raise(),
           out='optchi_p14_mx')),
     (dict(args=dict(kmm='kmm_ar', gammastar='gammastar_ar',
                     ns_star='ns_star_ar', ca='ca_ar', vpd='vpd_ar'),
           method='prentice14',  # scalar + arrays, c3
           cmng=does_not_raise(),
           out='optchi_p14_ar')),
     (dict(args=dict(kmm='kmm_ar', gammastar='shape_error',
                     ns_star='ns_star_ar', ca='ca_ar', vpd='vpd_ar'),
           method='prentice14',  # scalar + arrays, c3
           cmng=pytest.raises(ValueError),
           out=None))
     ]
)
def test_calc_optimal_chi(values, ctrl):

    if ctrl['method'] == 'c4':
        pytest.skip('Not currently testing C4 outputs - chi estimates for C4 need estimating as C3.')


    with ctrl['cmng']:
        kwargs = {k: values[v] for k, v in ctrl['args'].items()}
        ret = pmodel.CalcOptimalChi(**kwargs, method=ctrl['method'])

        expected = values[ctrl['out']]
        assert np.allclose(ret.chi, expected['chi'])
        assert np.allclose(ret.mj, expected['mj'])
        assert np.allclose(ret.mc, expected['mc'])
        assert np.allclose(ret.mjoc, expected['mjoc'])

# ------------------------------------------
# Testing CalcLUEVcmax -  This has quite a few combinations:
# - c4
# - optchi - using both scalar and array inputs.
# - soilmstress - only considering a scalar input here (so a uniform soil stress
#       across array optchi variants) rather than creating tests that have a
#       scalar optchi applied across array of soil moistures.
# - ftemp_kphio
# - CalcLUEVcmax method
# - kphio - normally this varies with the pmodel input setup but imposing a
#       single value here (0.05) to simplify test suite.
# ------------------------------------------


@pytest.mark.parametrize(
    'soilmstress',
    [dict(soilm=None, meanalpha=None),
     dict(soilm='soilm_sc', meanalpha='meanalpha_sc')],
    ids=['sm-off', 'sm-on']
)
@pytest.mark.parametrize(
    'ftemp_kphio',
    [True, False],
    ids=['fkphio-on', 'fkphio-off']
)
@pytest.mark.parametrize(
    'luevcmax_method',
    ['wang17', 'smith19', 'none'],
    ids=['wang17', 'smith19', 'none']
)
@pytest.mark.parametrize(
    'optchi',
    [dict(kmm='kmm_sc', gammastar='gammastar_sc',
          ns_star='ns_star_sc', ca='ca_sc', vpd='vpd_sc'),
     dict(kmm='kmm_ar', gammastar='gammastar_ar',
          ns_star='ns_star_ar', ca='ca_ar', vpd='vpd_ar')],
    ids=['sc', 'ar']
)
@pytest.mark.parametrize(
    'c4',
    [True],  # , False],
    ids=['c4']  # , 'c3']
)
def test_calc_lue_vcmax(request, values, soilmstress, ftemp_kphio,
                        luevcmax_method, optchi, c4):

    # ftemp_kphio needs to know the original tc inputs to optchi - these have
    # all been synchronised so that anything with type 'mx' or 'ar' used the
    # tc_ar input

    if c4:
        oc_method = 'c4'
        pytest.skip('Not currently testing C4 outputs because of param in rpmodel')
    else:
        oc_method = 'prentice14'

    if not ftemp_kphio:
        ftemp_kphio = 1.0
    elif optchi['kmm'] == 'kmm_sc':
        ftemp_kphio = pmodel.calc_ftemp_kphio(tc=values['tc_sc'], c4=c4)
    else:
        ftemp_kphio = pmodel.calc_ftemp_kphio(tc=values['tc_ar'], c4=c4)

    # Optimal Chi
    kwargs = {k: values[v] for k, v in optchi.items()}
    optchi = pmodel.CalcOptimalChi(**kwargs, method=oc_method)

    # Soilmstress
    if soilmstress['soilm'] is None:
        soilmstress = 1.0
    else:
        soilmstress = pmodel.calc_soilmstress(soilm=values[soilmstress['soilm']],
                                              meanalpha=values[soilmstress['meanalpha']])

    ret = pmodel.CalcLUEVcmax(optchi, kphio=0.05, ftemp_kphio=ftemp_kphio,
                              soilmstress=soilmstress, method=luevcmax_method)

    # Find the expected values, extracting the combination from the request
    name = request.node.name
    name = name[(name.find('[') + 1):-1]
    expected = values['jmax-' + name]

    assert np.allclose(ret.lue, expected['lue'])
    assert np.allclose(ret.vcmax, expected['vcmax_unitiabs'])

    if luevcmax_method == 'smith19':
        assert np.allclose(ret.omega, expected['omega'])
        assert np.allclose(ret.omega_star, expected['omega_star'])

# ------------------------------------------
# Testing PModelEnvironment class
# - double checking that the bundled calcs in here work as expected
# - test that the constraint issues a warning as expected.
# ------------------------------------------

@pytest.mark.parametrize(
    'variables',
    [dict(tc='tc_sc', vpd='vpd_sc', co2='co2_sc', patm='patm_sc',
          ca='ca_sc', kmm='kmm_sc', gammastar='gammastar_sc', ns_star='ns_star_sc'),
     dict(tc='tc_ar', vpd='vpd_ar', co2='co2_ar', patm='patm_ar',
          ca='ca_ar', kmm='kmm_ar', gammastar='gammastar_ar', ns_star='ns_star_ar')],
    ids=['sc', 'ar']
)
def test_pmodelenvironment(values, variables):

    ret = pmodel.PModelEnvironment(tc=values[variables['tc']],
                                   patm=values[variables['patm']],
                                   vpd=values[variables['vpd']],
                                   co2=values[variables['co2']]
                                   )

    assert np.allclose(ret.gammastar, values[variables['gammastar']])
    assert np.allclose(ret.ns_star, values[variables['ns_star']])
    assert np.allclose(ret.kmm, values[variables['kmm']])
    assert np.allclose(ret.ca, values[variables['ca']])


def test_pmodelenvironment_constraint():

    with pytest.warns(UserWarning):
        ret = pmodel.PModelEnvironment(tc=np.array([-15, 5, 10, 15, 20]),
                                       vpd=100000,
                                       co2=400,
                                       patm=101325)


def test_pmodelenvironment_toocold():

    with pytest.raises(ValueError):
        ret = pmodel.PModelEnvironment(tc=np.array([-35, 5, 10, 15, 20]),
                                       vpd=1000,
                                       co2=400,
                                       patm=101325)


def test_pmodelenvironment_dewpoint():

    with pytest.raises(ValueError):
        ret = pmodel.PModelEnvironment(tc=np.array([-15, 5, 10, 15, 20]),
                                       vpd=-1,
                                       co2=400,
                                       patm=101325)


# ------------------------------------------
# Testing PModel class - separate c3 and c4 tests
# - sc + ar inputs: tc, vpd, co2, patm (not testing elev)
# - +- soilmstress: soilm, meanalpha (but again assuming constant across ar inputs)
# - do_ftemp_kphio
# - luevcmax method

# For all tests
# - include fapar_sc and ppfd_sc (same irradiation everywhere)
# - hold kphio static
# ------------------------------------------


@pytest.fixture(scope='module')
def pmodelenv(values):
    """Fixture to create PModelEnvironments with scalar and array inputs
    """

    sc = pmodel.PModelEnvironment(tc=values['tc_sc'],
                                  vpd=values['vpd_sc'],
                                  co2=values['co2_sc'],
                                  patm=values['patm_sc'])

    ar = pmodel.PModelEnvironment(tc=values['tc_ar'],
                                  vpd=values['vpd_ar'],
                                  co2=values['co2_ar'],
                                  patm=values['patm_ar'])

    return {'sc': sc, 'ar': ar}


@pytest.mark.parametrize(
    'soilmstress',
    [False, True],
    ids=['sm-off', 'sm-on']
)
@pytest.mark.parametrize(
    'ftemp_kphio',
    [True, False],
    ids=['fkphio-on', 'fkphio-off']
)
@pytest.mark.parametrize(
    'luevcmax_method',
    ['wang17', 'smith19', 'none'],
    ids=['wang17', 'smith19', 'none']
)
@pytest.mark.parametrize(
    'environ',
    ['sc', 'ar'],
    ids=['sc', 'ar']
)
def test_pmodel_class_c3(request, values, pmodelenv, soilmstress, ftemp_kphio, luevcmax_method, environ):


    if soilmstress:
        soilmstress = pmodel.calc_soilmstress(values['soilm_sc'], values['meanalpha_sc'])
    else:
        soilmstress = None

    ret = pmodel.PModel(pmodelenv[environ],
                        kphio=0.05,
                        soilmstress=soilmstress,
                        do_ftemp_kphio=ftemp_kphio,
                        method_jmaxlim=luevcmax_method)

    # Find the expected values, extracting the combination from the request
    name = request.node.name
    name = name[(name.find('[') + 1):-1]
    expected = values['rpmodel-c3-' + name + '-unitiabs']

    # Test values - two values calculated in main rpmodel function
    # so can only test here - ci and iwue
    assert np.allclose(ret.iwue, expected['iwue'])
    assert np.allclose(ret.optchi.ci, expected['ci'])

    # - and six values that are scaled by IABS - rpmodel enforces scaling
    # where PModel can do it post hoc from unit_iabs values, so two
    # rpmodel runs are used to test the unit values and scaled.
    assert np.allclose(ret.unit_iabs.lue, expected['lue'])
    assert np.allclose(ret.unit_iabs.vcmax, expected['vcmax'])
    assert np.allclose(ret.unit_iabs.vcmax25, expected['vcmax25'])
    assert np.allclose(ret.unit_iabs.rd, expected['rd'])
    assert np.allclose(ret.unit_iabs.gs, expected['gs'])

    # TODO - Numerical instability in the Jmax calculation - as denominator
    #        approaches 1, results --> infinity unpredictably with rounding
    #        so currently excluding Jmax in combinations where this occurs.

    if 'none-fkphio-off-sm-off' not in name:
        assert np.allclose(ret.unit_iabs.jmax, expected['jmax'])
    else:
        warnings.warn('Skipping Jmax test for cases with numerical instability')

    # Check Iabs scaling
    iabs = ret.unit_iabs.scale_iabs(values['fapar_sc'], values['ppfd_sc'])

    # Find the expected values, extracting the combination from the request
    expected = values['rpmodel-c3-' + name + '-iabs']

    assert np.allclose(iabs.gpp, expected['gpp'])
    assert np.allclose(iabs.vcmax, expected['vcmax'])
    assert np.allclose(iabs.vcmax25, expected['vcmax25'])
    assert np.allclose(iabs.rd, expected['rd'])
    assert np.allclose(iabs.gs, expected['gs'])

    if 'none-fkphio-off-sm-off' not in name:
        assert np.allclose(iabs.jmax, expected['jmax'])
    else:
        warnings.warn('Skipping Jmax test for cases with numerical instability')


# Testing PModel class with C4

@pytest.mark.skipif(RPMODEL_C4_BUG, reason='Benchmark incorrect')
@pytest.mark.parametrize(
    'soilmstress',
    [False, True],
    ids=['sm-off', 'sm-on']
)
@pytest.mark.parametrize(
    'ftemp_kphio',
    [True, False],
    ids=['fkphio-on', 'fkphio-off']
)
@pytest.mark.parametrize(
    'environ',
    ['sc', 'ar'],
    ids=['sc', 'ar']
)
def test_pmodel_class_c3(request, values, pmodelenv, soilmstress, ftemp_kphio, environ):

    if soilmstress:
        soilmstress = pmodel.calc_soilmstress(values['soilm_sc'], values['meanalpha_sc'])
    else:
        soilmstress = None

    ret = pmodel.PModel(pmodelenv[environ],
                        kphio=0.05,
                        soilmstress=soilmstress,
                        do_ftemp_kphio=ftemp_kphio,
                        method_jmaxlim='none',  # enforced in rpmodel.
                        c4=True)

    # Find the expected values, extracting the combination from the request
    name = request.node.name
    name = name[(name.find('[') + 1):-1]
    expected = values['rpmodel-c4-' + name + '-unitiabs']

    # Test values - two values calculated in main rpmodel function
    # so can only test here - ci and iwue
    assert np.allclose(ret.iwue, expected['iwue'])
    assert np.allclose(ret.optchi.ci, expected['ci'])

    # - and six values that are scaled by IABS - rpmodel enforces scaling
    # where PModel can do it post hoc from unit_iabs values, so two
    # rpmodel runs are used to test the unit values and scaled.

    ret.estimate_productivity()  # defaults of fapar=1, ppfd=1

    assert np.allclose(ret.lue, expected['lue'])
    assert np.allclose(ret.vcmax, expected['vcmax'])
    assert np.allclose(ret.vcmax25, expected['vcmax25'])
    assert np.allclose(ret.rd, expected['rd'])
    assert np.allclose(ret.jmax, expected['jmax'])
    assert np.allclose(ret.gs, expected['gs'])

    # Check Iabs scaling
    ret.estimate_productivity(fapar=values['fapar_sc'],
                              ppfd=values['ppfd_sc'])

    # Find the expected values, extracting the combination from the request
    expected = values['rpmodel-c4-' + name + '-iabs']

    assert np.allclose(ret.gpp, expected['gpp'])
    assert np.allclose(ret.vcmax, expected['vcmax'])
    assert np.allclose(ret.vcmax25, expected['vcmax25'])
    assert np.allclose(ret.rd, expected['rd'])
    assert np.allclose(ret.jmax, expected['jmax'])
    assert np.allclose(ret.gs, expected['gs'])



from typing import Union, Tuple
import numpy as np

from pyrealm.utilities import check_input_shapes
from pyrealm.param_classes import TModelTraits

# Design Notes:
#
# One functionally easy thing to do the TTree object is to expose the geometry
# methods (e.g. by having TTree.calculate_crown_area()) to allow users to
# subclass the object and substitute their own geometry. Which is neat but the
# geometry and growth are not separate and there is no guarantee that a
# user-supplied geometry behaves as expected. So, easy to implement and a
# natural solution to people wanting to tinker with the geometry but not going
# there now.


class TTree:
    """Implementation of the T model

    This class provides an implementation of the calculations of tree geometry,
    mass and growth described by :cite:`Li:2014bc`. All of the properties of
    the T model are derived from a set of traits (see :class:`~pyrealm.tmodel.Traits`),
    stem diameter measurements and estimates of gross primary productivity.

    See the details of :meth:`~pyrealm.tmodel.TTree.set_diameter` and
    :meth:`~pyrealm.tmodel.TTree.calculate_growth` for details of the properties
    and calculations.

    Args:
        traits: An object of class :class:`~pyrealm.param_classes.TModelTraits`
    """

    def __init__(self, traits: TModelTraits = TModelTraits()):

        self.traits = traits

        # The diameter is used to define all of the geometric scaling
        # based on the trait parameters. It is set by the set_diameter()
        # method, which then populates the other geometric variables
        self._diameter = None
        self._height = None
        self._crown_fraction = None
        self._crown_area = None
        self._mass_stm = None
        self._mass_fol = None
        self._mass_swd = None

        # Growth is then applied by providing estimated gpp using the
        # calculate_growth() method, which populates the following:
        self._gpp_raw= None
        self._gpp_actual = None
        self._npp = None
        self._resp_swd = None
        self._resp_frt = None
        self._resp_fol = None
        self._turnover = None
        self._d_mass_s = None
        self._d_mass_fr = None
        self._delta_d = None
        self._delta_mass_stm = None
        self._delta_mass_frt = None


    @property
    def diameter(self):
        return self._diameter

    @property
    def height(self):
        return self._height

    @property
    def crown_fraction(self):
        return self._crown_fraction

    @property
    def crown_area(self):
        return self._crown_area

    @property
    def mass_swd(self):
        return self._mass_swd

    @property
    def mass_stm(self):
        return self._mass_stm

    @property
    def mass_fol(self):
        return self._mass_fol

    @property
    def gpp_raw(self):
        return self._gpp_raw

    @property
    def gpp_actual(self):
        return self._gpp_actual

    @property
    def resp_swd(self):
        return self._resp_swd

    @property
    def resp_frt(self):
        return self._resp_frt

    @property
    def resp_fol(self):
        return self._resp_fol

    @property
    def npp(self):
        return self._npp

    @property
    def turnover(self):
        return self._turnover

    @property
    def d_mass_s(self):
        return self._d_mass_s

    @property
    def d_mass_fr(self):
        return self._d_mass_fr

    @property
    def delta_d(self):
        return self._delta_d

    @property
    def delta_mass_stm(self):
        return self._delta_mass_stm

    @property
    def delta_mass_frt(self):
        return self._delta_mass_frt

    def set_diameter(self, values: Union[float, np.ndarray]):
        """Set stem diameter for the T model

        The set_diameter method sets the diameter values and then uses these
        values to populate a geometric and mass properties that scale with
        stem diameter.

        * Height (m, ``height``, :math:`H`):

        .. math::

            H = H_{max} ( 1 - e^{a D/ H_{max}}

        Returns:

        """

        self._diameter = values

        # Height of tree from diameter, Equation (4) of Li ea.
        self._height = (self.traits.h_max *
                        (1 - np.exp(-self.traits.a_hd * self.diameter
                                    / self.traits.h_max)))

        # Crown area of tree, Equation (8) of Li ea.
        self._crown_area = (((np.pi * self.traits.ca_ratio)/(4 * self.traits.a_hd))
                            * self.diameter * self.height)

        # Crown fraction, Equation (11) of Li ea.
        self._crown_fraction = self.height / (self.traits.a_hd * self.diameter)

        # Masses
        self._mass_stm = ((np.pi / 8) * (self.diameter ** 2) *
                          self.height * self.traits.rho_s)
        self._mass_fol = (self.crown_area * self.traits.lai *
                          (1 / self.traits.sla))
        self._mass_swd = (self.crown_area * self.traits.rho_s * self.height *
                          (1 - self.crown_fraction / 2) / self.traits.ca_ratio)

        # Clear any calculated growth values
        self._gpp_raw = None
        self._gpp_actual = None
        self._npp = None
        self._resp_swd = None
        self._resp_frt = None
        self._resp_fol = None
        self._turnover = None
        self._d_mass_s = None
        self._d_mass_fr = None
        self._delta_d = None
        self._delta_mass_stm = None
        self._delta_mass_frt = None

    def calculate_growth(self, gpp):
        """
        Grows a tree given estimated primary productivity.

        Args:
            gpp: Primary productivity

        Returns:

        """

        # GPP fixed per m2 of crown
        self._gpp_raw = gpp
        gpp_unit_cr = self.gpp_raw * (1 - np.exp(-(self.traits.par_ext * self.traits.lai)))
        self._gpp_actual = self.crown_area * gpp_unit_cr

        # Respiration costs (Eqn 13 of Li ea)
        # - sapwood, fine root and foliage maintenance
        self._resp_swd = self.mass_swd * self.traits.resp_s
        self._resp_frt = (self.traits.zeta * self.traits.sla *
                          self.mass_fol * self.traits.resp_r)
        self._resp_fol = (self.gpp_actual * self.traits.resp_f)

        # Net primary productivity
        self._npp = self.traits.yld * (self.gpp_actual - self.resp_fol -
                                       self.resp_frt - self._resp_swd)

        # Turnover costs for foliage and fine roots
        self._turnover = (self.crown_area * self.traits.lai *
                          ((1 / (self.traits.sla * self.traits.tau_f)) +
                           (self.traits.zeta / self.traits.tau_r)))

        # relative increments - these are used to calculate delta_d and
        # then scaled by delta_d to give actual increments
        self._d_mass_s = (np.pi / 8 * self.traits.rho_s * self.diameter *
                          (self.traits.a_hd * self.diameter *
                           (1 - (self.height / self.traits.h_max)) + 2 * self.height))

        self._d_mass_fr = (self.traits.lai *
                           ((np.pi * self.traits.ca_ratio) / (4 * self.traits.a_hd)) *
                           (self.traits.a_hd * self.diameter *
                            (1 - self.height / self.traits.h_max) + self.height) *
                           (1 / self.traits.sla + self.traits.zeta))

        # Actual increments
        self._delta_d = (self.npp - self.turnover) / (self.d_mass_s + self.d_mass_fr)
        self._delta_mass_stm = self.d_mass_s * self.delta_d
        self._delta_mass_frt = self.d_mass_fr * self.delta_d


def grow_ttree(gpp: Union[float, np.ndarray], d_init: Union[float, np.ndarray],
               time_axis: int, traits: TModelTraits = TModelTraits(),
               outvars: Tuple[str, ...] = ('diameter', 'height', 'crown_area', 'delta_d')):

    """Fits the T Model given a GPP

    Args:
        gpp: An array of GPP values
        d_init: An array of starting diameters
        traits: Traits to be used
        time_axis: An axis in P0 and d that represents an annual time series
        outvars: A list of Tree properties to store.

    Returns:

    """

    # The gpp array should contain a time axis to increment over. The d_init input
    # should then be the same shape as P0 _without_ that time axis dimension
    # and the loop will iterate over the years

    gpp_shape = gpp.shape
    if time_axis < 0 or time_axis >= len(gpp_shape):
        raise RuntimeError(f'time_axis must be >= 0 and <= {len(gpp_shape) - 1}')

    # Check that the input shapes for a single year match the shape of the
    # initial diameters
    single_year_gpp = np.take(gpp, indices=0, axis=time_axis)
    _ = check_input_shapes(single_year_gpp, d_init)

    # TODO - handle 1D GPP time series applied to more than one diameter

    # Initialise the Tree object
    tree = TTree(traits)

    # Check the requested outvars
    if 'diameter' not in outvars:
        raise RuntimeWarning('output_vars must include diameter')

    badvars = []
    for var in outvars:

        try:
            _ = getattr(tree, var)
        except AttributeError:
            badvars.append(var)

    if badvars:
        raise RuntimeError(f"Unknown tree properties in outvars: {', '.join(badvars)}")

    # Create an array to store the requested variables by adding a dimension
    # to the gpp input with length set to the number of variables
    output_shape = gpp_shape + tuple([len(outvars)])
    output = np.zeros(output_shape)

    # Insert the initial diameters
    tree.set_diameter(d_init)

    # Create an indexing object to insert values into the output. This is
    # a bit obscure: the inputs have a time axis and an arbitrary number
    # of other dimensions and the output adds another dimension for the
    # output variables. So this object creates a slice (:) on _all_ dimensions
    # and the loop then replaces the time axis and variable axis with integers.
    output_index = [slice(None)] * output.ndim

    # Loop over the gpp time axis
    for year in np.arange(gpp_shape[time_axis]):

        # Calculate the growth based on the current year of gpp
        tree.calculate_growth(np.take(gpp, indices=year, axis=time_axis))

        # Store the requested variables into the output array
        for var_idx, each_var in enumerate(outvars):

            # Extract variable values from tree into output - set the last index
            # (variable dimension) to the variable index and the time axis to the year
            output_index[-1] = var_idx
            output_index[time_axis] = year
            output[tuple(output_index)] = getattr(tree, each_var)

        # Now update the tree object
        tree.set_diameter(tree.diameter + getattr(tree, 'delta_d'))

    return output






import os
import netCDF4
from pyrealm import pmodel
import numpy as np
import pytest


@pytest.fixture(scope='module')
def dataset():
    """Fixture to load test inputs from file from data folder in package root
    """
    
    test_dir = os.path.dirname(os.path.abspath(__file__))
    data_file_path = os.path.normpath(os.path.join(test_dir, os.pardir, os.pardir, 'data', 'pmodel_inputs.nc'))

    dataset = netCDF4.Dataset(data_file_path)
    
    # TODO - masked arrays should be handled (it is basically the same as
    #        NA values in the R implementation) but for current validation
    #        this is just going to convert them to filled arrays

    dataset.set_auto_mask(False)

    # Extract the six variables for all months
    temp = dataset['temp'][:]
    co2 = dataset['CO2'][:]         # Note - spatially constant but mapped.
    elev = dataset['elevation'][:]  # Note - temporally constant but repeated
    vpd = dataset['VPD'][:]
    fapar = dataset['fAPAR'][:]
    ppfd = dataset['ppfd'][:]

    dataset.close()

    # Convert elevation to atmospheric pressure
    patm = pmodel.calc_patm(elev)

    # Calculate p model environment
    env = pmodel.PModelEnvironment(tc=temp, vpd=vpd, co2=co2, patm=patm)

    return env, fapar, ppfd

@pytest.mark.skipif(True, reason='Broken implementation in rpmodel 1.2.0')
@pytest.mark.parametrize(
    'ctrl',
    [(dict(rfile='rpmodel_global_gpp_no_ftkphio.nc',
           do_ftemp_kphio=False)),
     (dict(rfile='rpmodel_global_gpp_do_ftkphio.nc',
           do_ftemp_kphio=True))]
)
def test_pmodel_global_array(dataset, ctrl):
    """This test runs a comparison between the rpmodel outputs and
    pyrealm.pmodel using global data at 0.5 degree resolution. The input
    data file contains values of the key variables for 2 months, giving
    three dimensional inputs. The same data have been run through rpmodel
    using the file test_global_array.R and the resulting GPP stored in
    rpmodel_global_gpp.nc
    """

    env, fapar, ppfd = dataset

    # Skip ftemp kphio is false
    if ctrl['do_ftemp_kphio']:
        pytest.skip('Not currently testing global array against rpmodel with ftemp_kphio due to bug pre 1.2.0')

    # Run the P model
    model = pmodel.PModel(env, kphio=0.05, do_ftemp_kphio=ctrl['do_ftemp_kphio'])

    # Scale the outputs from values per unit iabs to realised values
    scaled = model.unit_iabs.scale_iabs(fapar, ppfd)

    # Load the R outputs
    test_dir = os.path.dirname(os.path.abspath(__file__))

    ds = netCDF4.Dataset(os.path.join(test_dir, ctrl['rfile']))
    gpp_r = ds['gpp'][:]

    # # Debugging code:
    # # Export the pyrealm values to a new netcdf file using the R one as a template
    # with netCDF4.Dataset("pyrealm_global_gpp.nc", "w", format="NETCDF3_CLASSIC") as gpp:
    #
    #     for name, dimension in ds.dimensions.items():
    #         gpp.createDimension(
    #             name, (len(dimension) if not dimension.isunlimited() else None))
    #
    #     # copy all file data except for the excluded
    #     for name, variable in ds.variables.items():
    #
    #         gpp.createVariable(name, variable.datatype, variable.dimensions)
    #
    #         # copy variable attributes all at once via dictionary
    #         gpp[name].setncatts(ds[name].__dict__)
    #
    #         # input data - truncate to the first two time slots
    #         if name == 'gpp':
    #             gpp[name][:] = scaled.gpp
    #         else:
    #             gpp[name][:] = ds[name][:]
    #
    # ds.close()

    assert scaled.gpp.shape == gpp_r.shape
    assert np.allclose(scaled.gpp, gpp_r, equal_nan=True)

    # ## Run the P model in a location where the trimming in do_ftemp_kphio matters
    # patm = pmodel.calc_patm(1211)
    # model = pmodel.PModel(tc=-25.5, co2=390.2, patm=patm, vpd=91.77804,
    #                       kphio=0.05, do_ftemp_kphio=False)
    # scaled = model.unit_iabs.scale_iabs(fapar=0.3468055, ppfd=74848.94)
    #
    # ## Run the P model in a loction with missing data
    #
    # idx = (0, 176, 653)
    #
    # patm = pmodel.calc_patm(elev[idx])
    # model = pmodel.PModel(tc=temp[idx], co2=co2[idx], patm=patm, vpd=vpd[idx],
    #                       kphio=0.05, do_ftemp_kphio=False)
    # scaled = model.unit_iabs.scale_iabs(fapar=fapar[idx], ppfd=ppfd[idx])




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

version = {}
with open("pyrealm/version.py") as fp:
    exec(fp.read(), version)

setuptools.setup(
    name="pyrealm",
    version=version['__version__'],
    author="David Orme",
    author_email="d.orme@imperial.ac.uk",
    description="Python implementations of REALM models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://pyrealm.readthedocs.io/",
    packages=setuptools.find_packages(),
    # include_package_data=True,
    # package_data={'': ['data/*.nc']},
    entry_points={
            'console_scripts': ['']
    },
    license='MIT',
    install_requires=['dacite',
                      'numpy',
                      'enforce_typing',
                      'tabulate'],
    python_requires='>=3.6',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Development Status :: 3 - Alpha"
    ],
)
